#include "MainWindow.hpp"
#include <Shlwapi.h>
#include "constants.h"
#include "KeyBind.hpp"
#include "KeyboardSynthesizer.hpp"
#include "MidiSynthesizer.hpp"
#include "KeySequencer.hpp"
#include "KeyBindWindow.hpp"
#include "resource.h"
#include "StringTable.hpp"
#include "MidiInput.hpp"

MainWindow::MainWindow(HINSTANCE instanceHandle)
: initialized_(false), instanceHandle_(instanceHandle), windowHandle_(0), fileTextHandle_(0), statusTextHandle_(0), keySequencer_(nullptr)
{
	static int staticInitialize = registerWindowClass(instanceHandle);

	keyBind_ = new KeyBind();
	keyboardSynthesizer_ = new KeyboardSynthesizer(keyBind_);
	midiSynthesizer_ = new MidiSynthesizer();
	midiInput_ = new MidiInput(keyboardSynthesizer_);
}

MainWindow::~MainWindow()
{
	delete midiInput_;
	delete midiSynthesizer_;
	delete keyboardSynthesizer_;
	delete keyBind_;
}

bool MainWindow::initialize()
{
	if (initialized_)
		return false;
	initialized_ = true;

	// 쐬EBhẼX^C
	DWORD windowStype = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;

	// NCAg̈悩EBhETCYvZ
	int width = CLIENT_WIDTH;
	int height = CLIENT_HEIGHT;
	RECT clientRect = { 0, 0, width, height };
	if (AdjustWindowRect(&clientRect, windowStype, FALSE))
	{
		width = clientRect.right - clientRect.left;
		height = clientRect.bottom - clientRect.top;
	}

	// ʒʒuƂ
	int x = GetSystemMetrics(SM_CXSCREEN);
	int y = GetSystemMetrics(SM_CYSCREEN);
	x = x > width ? (x - width) / 2 : 0;
	y = y > height ? (y - height) / 2 : 0;

	// Window쐬
	windowHandle_ = CreateWindowEx(0, WINDOW_CLASS_NAME, StringTable::get(IDS_MAINWINDOW_CAPTION), windowStype, x, y, width, height, nullptr, nullptr, instanceHandle_, nullptr);
	if (windowHandle_ == nullptr)
	{
		MessageBox(nullptr, StringTable::get(IDS_ERROR_CREATE_WINDOW), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
		return false;
	}

	// Window procedurepthis|C^L
	SetWindowLongPtr(windowHandle_, GWLP_USERDATA, 0);
	if (SetWindowLongPtr(windowHandle_, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this)) == 0 && GetLastError() != 0)
		goto DESTROY;

	// Drag & DropL
	DragAcceptFiles(windowHandle_, TRUE);

	WORD hotkey = keyBind_->getPlayHotKey();
	if (hotkey != 0)
	{
		UINT modifiers = 0;
		if (hotkey & KeyBind::ModifierKey::SHIFT)
			modifiers |= MOD_SHIFT;
		if (hotkey & KeyBind::ModifierKey::CONTROL)
			modifiers |= MOD_CONTROL;
		if (hotkey & KeyBind::ModifierKey::ALTERNATE)
			modifiers |= MOD_ALT;

		if (RegisterHotKey(windowHandle_, ID_HOTKEY_PLAY, modifiers, hotkey & 0x00ff) == 0)
		{
			MessageBox(windowHandle_, StringTable::get(IDS_ERROR_REGISTER_PLAY_HOTKEY), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
		}
	}

	// Đt@C\
	fileTextHandle_ = CreateWindow(_T("STATIC"), StringTable::get(IDS_MAINWINDOW_FILE_EMPTY), WS_CHILD | WS_VISIBLE | WS_BORDER | SS_CENTERIMAGE | SS_LEFT, 8, 8, CLIENT_WIDTH - 16, 30, windowHandle_, nullptr, instanceHandle_, nullptr);
	if (fileTextHandle_ == nullptr)
		goto DESTROY;

	// ĐXe[^X\
	statusTextHandle_ = CreateWindow(_T("STATIC"), _T(""), WS_CHILD | WS_VISIBLE | WS_BORDER | SS_CENTERIMAGE | SS_CENTER, CLIENT_WIDTH - 100 - 8, 46, 100, 30, windowHandle_, nullptr, instanceHandle_, nullptr);
	if (statusTextHandle_ == nullptr)
		goto DESTROY;

	// L[oChݒ{^\
	openKeyBindButtonHanndle_ = CreateWindow(_T("BUTTON"), StringTable::get(IDS_MAINWINDOW_KEY_BIND_BUTTON), WS_CHILD | WS_VISIBLE | WS_BORDER | BS_PUSHBUTTON | BS_CENTER, 8, 46, 100, 30, windowHandle_, reinterpret_cast<HMENU>(ID_BUTTON_KEY_BIND), instanceHandle_, nullptr);
	if (openKeyBindButtonHanndle_ == nullptr)
		goto DESTROY;

	// eXgĐ/~{^\
	testPlayButtonHandle_ = CreateWindow(_T("BUTTON"), _T(""), WS_CHILD | WS_VISIBLE | WS_BORDER | BS_PUSHBUTTON | BS_CENTER, 116, 46, 192, 30, windowHandle_, reinterpret_cast<HMENU>(ID_BUTTON_TEST_PLAY), instanceHandle_, nullptr);
	if (testPlayButtonHandle_ == nullptr)
		goto DESTROY;

	// MIDI͂̃`FbN{bNX
	midiInputCheckboxHandle_ = CreateWindow(_T("BUTTON"), StringTable::get(IDS_MAINWINDOW_MIDI_INPUT_CHECKBOX), WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX, 8, 84, 256, 30, windowHandle_, reinterpret_cast<HMENU>(ID_CHECKBOX_MIDI_INPUT), instanceHandle_, nullptr);
	if (midiInputCheckboxHandle_ == nullptr)
		goto DESTROY;
	else
	{
		// .init@CMIDI͂̃`FbNԂǂݏo
		if (PathFileExists(INI_FILE_PATH))
		{
			TCHAR buffer[2];
			if (GetPrivateProfileString(_T("Configuration"), _T("MidiInput"), _T(""), buffer, 2, INI_FILE_PATH) == 1)
			{
				if (buffer[0] == _T('1'))
				{
					LRESULT result = midiInput_->open();
					if (result == MidiInput::ERROR_DEVICE_NOT_FOUND)
						MessageBox(windowHandle_, StringTable::get(IDS_ERROR_MIDI_INPUT_DEVICE_NOT_FOUND), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					else if (result == MidiInput::ERROR_UNKNOWN)
						MessageBox(windowHandle_, StringTable::get(IDS_ERROR_MIDI_INPUT_UNKNOWN), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					else
						SendMessage(midiInputCheckboxHandle_, BM_SETCHECK, 1, 0);
				}
			}
		}
	}

	// Window\
	ShowWindow(windowHandle_, SW_SHOW);

	// t쐬
	keySequencer_ = new KeySequencer(windowHandle_, keyBind_, keyboardSynthesizer_);
	keySequencer_->threadStart();

	// NɃt@ChbvĂǂݍ
	if (__argc > 1 && __targv[1] != nullptr)
	{
		LRESULT result = keySequencer_->loadFile(__targv[1]);
		if (result == 0)
		{
			// t@C̒擾
			WORD filenameLength = GetFileTitle(__targv[1], nullptr, 0);

			// t@Cp̃m
			TCHAR* filename = new TCHAR[filenameLength]; // ͏I[NULLlengthɓĂ

			// t@C擾
			GetFileTitle(__targv[1], filename, filenameLength);

			// t@CpX̕\؂ւ
			SetWindowText(fileTextHandle_, filename);

			delete[] filename;
		}
	}

	refreshStatus();

	return true;

DESTROY:

	MessageBox(windowHandle_, StringTable::get(IDS_ERROR_CREATE_WINDOW), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
	PostMessage(windowHandle_, WM_CLOSE, 0, 0);
	run();
	return false;
}

void MainWindow::run()
{
	MSG message;
	while (GetMessage(&message, nullptr, 0, 0) != 0)
		DispatchMessage(&message);
}

int MainWindow::registerWindowClass(HINSTANCE instanceHandle)
{
	// Ă㏑邾ővȋC邯
	// ꉞŏ1񂾂o^΂Ȃ̂ł̂悤ɏĂ
	// EBhENX̓o^
	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.lpszClassName = WINDOW_CLASS_NAME;
	wcex.hInstance = instanceHandle;
	wcex.style = 0;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.lpszMenuName = nullptr;
	wcex.hbrBackground = BACKGROUND_BRUSH;
	wcex.hIcon = (HCURSOR)LoadImage(instanceHandle, MAKEINTRESOURCE(IDI_ICON1), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE);
	wcex.hIconSm = (HCURSOR)LoadImage(instanceHandle, MAKEINTRESOURCE(IDI_ICON1), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE);
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.lpfnWndProc = callbackProcedure;
	RegisterClassEx(&wcex);

	return 0;
}

LRESULT WINAPI MainWindow::callbackProcedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
{
	MainWindow* window = reinterpret_cast<MainWindow*>(GetWindowLongPtr(windowHandle, GWLP_USERDATA));
	if (window != nullptr)
		return window->procedure(windowHandle, message, wParam, lParam);

	switch (message)
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	default:
		return DefWindowProc(windowHandle, message, wParam, lParam);
	}
}

LRESULT MainWindow::procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_APP_REFRESH_STATUS:
		refreshStatus();
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case ID_BUTTON_KEY_BIND:
			{
				// tȂ~߂
				stopKeySequencer();
				refreshStatus();
				
				KeyBindWindow* child = new KeyBindWindow(instanceHandle_, windowHandle, keyBind_);
				if (child->initialize())
					child->run();
				else
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_OPEN_KEYBINDWINDOW), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
				delete child;
			}
			return 0;

		case ID_BUTTON_TEST_PLAY:
			if (keySequencer_->isPlaying())
				stopKeySequencer();
			else
			{
				if (midiSynthesizer_->open())
				{
					keySequencer_->setSynthesizer(midiSynthesizer_);
					keySequencer_->play();
				}
				else
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_OPEN_MIDI_OUT), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
			}
			refreshStatus();
			return 0;

		case ID_CHECKBOX_MIDI_INPUT:
			if (BST_CHECKED == SendMessage(midiInputCheckboxHandle_, BM_GETCHECK, 0, 0))
			{
				LRESULT result = midiInput_->open();
				if (result == MidiInput::ERROR_DEVICE_NOT_FOUND)
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_MIDI_INPUT_DEVICE_NOT_FOUND), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
				else if (result == MidiInput::ERROR_UNKNOWN)
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_MIDI_INPUT_UNKNOWN), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
				else
				{
					// .init@Cɏ
					WritePrivateProfileString(_T("Configuration"), _T("MidiInput"), _T("1"), INI_FILE_PATH);
				}
			}
			else
			{
				midiInput_->close();

				// .init@Cɏ
				WritePrivateProfileString(_T("Configuration"), _T("MidiInput"), _T("0"), INI_FILE_PATH);
			}
			return 0;
		}
		break;

	case WM_HOTKEY:
		if (wParam == ID_HOTKEY_PLAY)
		{
			if (keySequencer_->isPlaying())
				stopKeySequencer();
			else
			{
				keySequencer_->setSynthesizer(keyboardSynthesizer_);
				keySequencer_->play();
			}
			refreshStatus();
		}
		return 0;

	case WM_DROPFILES:
		{
			HDROP handle = reinterpret_cast<HDROP>(wParam);

			// hbvꂽt@C(1)̃pX̒擾
			UINT filepathLength = DragQueryFile(handle, 0, nullptr, 0);

			// t@CpXp̃m
			TCHAR* filepath = new TCHAR[filepathLength + 1]; // I[NULLlengthɓȂ

			// t@CpX擾
			DragQueryFile(handle, 0, filepath, filepathLength + 1);
			DragFinish(handle);

			// tȂ~߂
			stopKeySequencer();
			refreshStatus();

			// t@Cǂݍ
			LRESULT result = keySequencer_->loadFile(filepath);
			if (result == 0)
			{
				// t@C̒擾
				WORD filenameLength = GetFileTitle(filepath, nullptr, 0);

				// t@Cp̃m
				TCHAR* filename = new TCHAR[filenameLength]; // ͏I[NULLlengthɓĂ

				// t@C擾
				GetFileTitle(filepath, filename, filenameLength);

				// t@CpX̕\؂ւ
				SetWindowText(fileTextHandle_, filename);

				delete[] filename;
			}
			else
			{
				SetWindowText(fileTextHandle_, StringTable::get(IDS_MAINWINDOW_FILE_EMPTY));

				switch (result)
				{
				case StandardMidiFormat::LOAD_ERROR_FILE_OPEN:
					// t@CI[voȂ
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_OPEN_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_FILE_TOO_LARGE:
					// 4GBȏ̃t@C͔Ή
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_TOO_LARGE_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_FILE_READ:
					// t@CǂݍݒɕsȃG[
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_READ_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_ILLEGAL_SMF:
					// SMFƂĕs
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_ILLEGAL_SMF_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_NOT_SUPPORT_SMF:
					// T|[gĂȂSMF`
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_UNSUPPORT_SMF_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				}
			}

			delete[] filepath;
		}
		return 0;

	case WM_CTLCOLORSTATIC:
		SetBkMode(reinterpret_cast<HDC>(wParam), TRANSPARENT);
		return reinterpret_cast<LRESULT>(BACKGROUND_BRUSH);

	case WM_CLOSE:
		// EBhEQƂ̂ɉĂDestroyWindow
		UnregisterHotKey(windowHandle, ID_HOTKEY_PLAY);

		if (keySequencer_ != nullptr)
		{
			delete keySequencer_;
			keySequencer_ = nullptr;
		}

		DestroyWindow(windowHandle);
		return 0;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(windowHandle, message, wParam, lParam);
}

void MainWindow::refreshStatus()
{
	if (keySequencer_->isPlaying())
	{
		SetWindowText(statusTextHandle_, StringTable::get(IDS_MAINWINDOW_STATUS_PLAYING));
		SetWindowText(testPlayButtonHandle_, StringTable::get(IDS_MAINWINDOW_TEST_PLAY_BUTTON_PLAYING));
	}
	else
	{
		SetWindowText(statusTextHandle_, StringTable::get(IDS_MAINWINDOW_STATUS_STOPPED));
		SetWindowText(testPlayButtonHandle_, StringTable::get(IDS_MAINWINDOW_TEST_PLAY_BUTTON_STOPPED));
	}
}

void MainWindow::stopKeySequencer()
{
	// t~
	keySequencer_->stop();

	// MIDIo͂(XĂ牽Ȃv)
	midiSynthesizer_->close();

	// o͂L[{[hɃZbg
	keySequencer_->setSynthesizer(keyboardSynthesizer_);
}